Canvas = class(Group);

function Canvas:Constructor()
    Group.Constructor(self, "Turbine.UI.Lotro.Window");
    self.idStr = "new Canvas";

    self.selectedShapes = {};

    -- Hardcoded properties
    self.properties.Class = Canvas;
    self.changedProperties.Class = Canvas;
    self.resizable = false; -- resizable, but not with a GrabPoint
    self.baseClass.SetResizable(self, true);
    self.baseClass.SetVisible(self, true);
    self.baseClass.SetZOrder(self, -1);

    -- Default properties
    self.applyingRemoteChanges = true;
    self:SetSize(620, 360);
    self.applyingRemoteChanges = false;
    
    self:SavePropertyDefaults();
end

function Canvas:_SizeChanged()
    -- Nothing to be done.
    -- Perhaps later: Adjust child-clipping, position of children, scrollbar positions
end

function Canvas:MouseMove()
    -- Override Shape:MouseMove() because Lotro.Window has built-in drag-movement functionality.
    -- Later: Add multi-selecting by drawing a rectangle around shapes.
end

function Canvas:AddShapeAt(shape, left, top)
    Group.AddShapeAt(self, shape, left, top);
    shape.prevLeft, shape.prevTop = left or 0, top or 0;
    AddCallback(shape, "SizeChanged", function(s)
        self:ShapeSizeChanged(s);
    end);
    AddCallback(shape, "PositionChanged", function(s)
        self:ShapePositionChanged(s);
        s.prevLeft, s.prevTop = s:GetPosition();
    end);
    AddCallback(shape, "MouseClick", function(s, args)
        if (s.mouseMovedAfterMouseDown) then
            -- Not a real "click"; we shouldn't have gotten this event.
            return;
        end
        if (not self:IsShiftKeyDown()) then
            self:Deselect();
        end
-- to do: check for ctrl key?
        self:SelectShape(s);
    end);
end

function Canvas:RemoveShape(shape)
    if (self.selectedShapes[shape]) then
        self:Deselect(shape);
    end
    Group.RemoveShape(self, shape);
end

function Canvas:CloneShape(shape)
    Group.CloneShape(self, shape);
    self:Reselect();
end

function Canvas:ApplyChanges(properties, from)
    Shape.ApplyChanges(self, properties, from);

    -- Selection may have changed.
    self:Reselect();
end

function Canvas:Reselect()
    -- Scan all shapes for IsSelected(), reapply their selection, and reconstruct the self.selectedShapes table.
    self.selectedShapes = {};
    for _, shape in pairs(self.shapesByIDStr) do
        if (shape:IsSelected()) then
            shape:SetSelected(true);
            self.selectedShapes[shape] = true;
        else
            shape:SetSelected(false);
            self.selectedShapes[shape] = nil;
        end
    end
    DoCallbacks(self, "SelectionChanged");
end

function Canvas:Deselect(shape)
    if (not shape) then
        -- deselect all selected shapes
        for shape, _ in pairs(self.selectedShapes) do
            self:Deselect(shape);
        end
    else
        shape:SetSelected(false);
        self.selectedShapes[shape] = nil;
    end
    DoCallbacks(self, "SelectionChanged");
end

-- Adds a shape to the current selection and puts a marquee on it
function Canvas:SelectShape(shape)
    self.selectedShapes[shape] = true;
    shape:SetSelected(true);
    DoCallbacks(self, "SelectionChanged");
end

function Canvas:GetSelection()
    return self.selectedShapes;
end

function Canvas:ShapeIsSelected(shape)
    return self.selectedShapes[shape] or false;
end

function Canvas:ShowSelectionContextMenu()
    local firstShape = next(self.selectedShapes);
    if (firstShape) then
        local secondShape = next(self.selectedShapes, firstShape);
        if (secondShape) then
            -- Multiple shapes selected
            -- to do: show things like align, distribute, etc. that apply to multiple shapes
        else
            -- Single shape selected
            firstShape:GetContextMenu():ShowMenu();
        end
    end
end

function Canvas:SetShape(str, from)
    Group.SetShape(self, str, from);
    self:Reselect();
end

function Canvas:ShapeSizeChanged(shape)
    if (self:ShapeIsSelected(shape)) then
        local width, height = shape:GetSize();
        shape.marquee:SetSize(width + 2, height + 2);
    end
end

function Canvas:ShapePositionChanged(shape)
    if (self:ShapeIsSelected(shape)) then
        local left, top = shape:GetPosition();
        shape.marquee:SetPosition(left - 1, top - 1);
        
        -- If other shapes are also selected, move them by the same amount
        if (not shape.doingGroupMove) then
            local xShift = left - shape.prevLeft;
            local yShift = top - shape.prevTop
            for s, _ in pairs(self.selectedShapes) do
                if (s ~= shape) then
                    local x, y = s:GetPosition();
                    s.doingGroupMove = true;
                    s:SetPosition(x + xShift, y + yShift);
                end
            end
        end
        shape.doingGroupMove = false;
    end
end

-- When the user clicks on the background behind the shapes
function Canvas:MouseClick(args)
    self:Deselect();
    self:Focus();
end

function Canvas:ImportShapes(str)
    Group.ImportShapes(self, str);
    self:Reselect();
end

function Canvas:SetEditorEnabled(enabled)
    if (enabled) then
        if (not self.editor) then
            self.editor = Editor(self);
            return self.editor;
        end
    end
    self.editor:SetVisible(enabled);
end

function Canvas:Activate()
    self.baseClass.Activate(self);
end

function Canvas:SetVersion(version) -- plugin version used to create this Canvas, times 100
    self:SetProperty("Version", math.floor(version * 100));
end

function Canvas:GetVersion()
    return self.properties.Version / 100;
end

-- Standard tracked properties
function Canvas:_SizeChanged()
    local width, height = self:GetSize();
    self:SetProperty("Width", width);
    self:SetProperty("Height", height);
end

function Canvas:SetText(text) -- Canvas window title
    local title = text;
    if (self.properties.Owner ~= self.playerName) then
        text = text .. " - " .. self.properties.Owner;
    end
    self.baseClass.SetText(self, text);
    self:SetProperty("Text", title);
    DoCallbacks(self, "TextChanged");
end

function Canvas:GetText()
    return self.properties.Text;
end