Editor = class(Turbine.UI.Window);

function Editor:Constructor(canvas)
    Turbine.UI.Window.Constructor(self);
    self:SetVisible(true);
--self:SetBackColor(Turbine.UI.Color(0.5, 0.1, 0.1, 0.1));
    self:SetMouseVisible(false);

    self.buttonTray = ButtonTray();
    self.buttonTray:SetParent(self);
    AddCallback(self.buttonTray, "ButtonClicked", function(_, button)
        self:TrayButtonClicked(button);
    end);

    self.canvas = canvas;
    self:SetZOrder(self.canvas:GetZOrder());
    AddCallback(self.canvas, "SizeChanged", function()
        self:Redraw();
    end);
    AddCallback(self.canvas, "PositionChanged", function()
        self:Redraw();
    end);
    AddCallback(self.canvas, "SelectionChanged", function()
        self:SelectTools();
    end);
    AddCallback(self.canvas, "Changed", function()
        self.idleTime = Turbine.Engine.GetGameTime();
        if (self.canvas:IsChanged()) then
            self.sendButtonNeedsUpdate = true;
        end
    end);
    AddCallback(self.canvas, "Activated", function()
        self:SelectTools();
    end);

    self.propertyTray = PropertyTray();
    self.propertyTray:SetParent(self);
    AddCallback(self.propertyTray, "ChangeRequested", function(_, args)
        self:PropertyChangeRequested(args.Property, args.Value);
    end);

    self.sendButton = SendButton();
    self.sendButton:SetParent(self);
    AddCallback(self.sendButton, "TransferStarted", function()
        self:MaskCanvas(true);
        self.canvas:ResetChanges();
        self:HideTools();
    end);
    AddCallback(self.sendButton, "TransferComplete", function()
        self:MaskCanvas(false);
        self.canvas:SetZOrder(self:GetZOrder() + 1);
        self:SetMouseVisible(false);
        self:SelectTools();
    end);
    AddCallback(self.sendButton, "MouseEnter", function()
--to do: disable self until all data is sent
        self.shimmering:Remove(self.sendButton);
    end);
    
    self.settingsButton = SettingsButton();
    self.settingsButton:SetParent(self);
    AddCallback(self.settingsButton, "Click", function()
        self:ShowSettingsMenu();
    end);
    
    self.shimmering = Thurallor.UI.Shimmering();

    self.playerName = Turbine.Gameplay.LocalPlayer:GetInstance():GetName();
    self.shapes = {};
    self.changedShapes = {};
    self.selectedShapes = {};
    self.changed = false;
    self.idleTime = 0;

    self:Redraw();
    self:SetWantsUpdates(true);
end

function Editor:Redraw()
    local spacing = 5;
    local margin = (spacing + 32 + spacing - 2);
    local topMargin = margin - 16;
    local propTrayHeight = 41;

    local width, height = self.canvas:GetSize();
    local left, top = self.canvas:GetPosition();
    
    left = left - margin;
    width = width + 2 * margin;
    
    top = top - topMargin;
    height = height + topMargin + spacing + propTrayHeight;
    
    self:SetSize(width, height);
    self:SetPosition(left, top);

    self.buttonTray:SetPosition(spacing, spacing);
    self.buttonTray:SetSize(32, height - 2 * spacing);
    
    self.propertyTray:SetPosition(margin + spacing, height - spacing - propTrayHeight);
    self.propertyTray:SetSize(width - 2 * margin - 2 * spacing, propTrayHeight);

    self.sendButton:SetPosition(width - margin + spacing - 2, math.floor(0.5 + height / 2) - 16);
    self.settingsButton:SetPosition(self.sendButton:GetLeft(), self.propertyTray:GetTop() - margin - spacing);
end

function Editor:MaskCanvas(enable)
    if (enable and (not self.mask)) then
        local mask = Turbine.UI.Control();
        mask:SetParent(self.canvas);
        mask:SetSize(self.canvas:GetSize());
        mask:SetZOrder(2147483647 - 1); -- leave room for the tooltip in front
        mask:SetBackColor(Turbine.UI.Color(0.5, 0.25, 0.25, 0.25));
        mask:SetBackColorBlendMode(Turbine.UI.BlendMode.Overlay);
        Thurallor.UI.Tooltip("You must finish sending the changes before you can resume editing."):Attach(mask);
        self.mask = mask;
    elseif ((not enable) and self.mask) then
        self.mask:SetParent(nil);
        self.mask = nil;
    end
end

function Editor:HideTools()
    self.buttonTray:SetContext(nil);
    self.propertyTray:SetContext(nil);
end

function Editor:SelectTools()
    local selectedShapes = self.canvas:GetSelection();
--Puts("select tools: " .. Serialize(selectedShapes));
    if (not next(selectedShapes)) then
        self.buttonTray:SetContext("Global");
        self.propertyTray:SetContext("Global");
    else
        local shapes = {};
        for shape, _ in pairs(selectedShapes) do
            local class = shape:GetClass();
            if (not shapes[class]) then
                shapes[class] = { shape };
            else
                table.insert(shapes[class], shape);
            end
        end
        self.buttonTray:SetContext("Shape", shapes);
        self.propertyTray:SetContext("Shape", shapes);
    end
end

function Editor:TrayButtonClicked(button)
    local shape;

    if (button == "CreateText") then
        shape = self:CreateTextShape();
    elseif (button == "CreateImage") then
        shape = self:CreateImageShape();
    elseif (button == "CreateGeometric") then
        shape = self:CreateRectangleShape();
    elseif (button == "CreateMap") then
        shape = self:CreateRadarMapShape();
    elseif (button == "CreateMap") then
        shape = self:CreateRadarMapShape();
    end
    
    -- If a new shape was created:
    if (shape) then
        self.canvas:Deselect();
        self.canvas:SelectShape(shape);
    end
end

function Editor:ShowSettingsMenu()
    local menu, items, item;
    menu = Turbine.UI.ContextMenu();
    items = menu:GetItems();

    item = Turbine.UI.MenuItem("Load...", false);
    items:Add(item);
    
    item = Turbine.UI.MenuItem("Save...", false);
    items:Add(item);

    item = Turbine.UI.MenuItem("Import...");
    item.Click = function()
        self:ShowImportWindow();
    end
    items:Add(item);

    item = Turbine.UI.MenuItem("Export...");
    item.Click = function()
        self:ExportCanvas();
    end
    items:Add(item);
    
    menu:ShowMenu();
end

function Editor:ExportCanvas()
    local title = "Exporting “" .. self.canvas:GetText() .. "”";
    local window = Alert(title, self.canvas:ExportShapes(), "Ok", Turbine.UI.Lotro.Font.Verdana14);
    CenterWindow(window);
    window:SetZOrder(2147483647);
    local okButton = window.buttons["Ok"];
    okButton.Click = function()
        self.exportWindow:Close();
    end
    self.exportWindow = window;
end

function Editor:ShowImportWindow()
    if (self.importWindow) then
        self.importWindow:Close();
    end
    local window = Turbine.UI.Lotro.Window();
    window:SetVisible(true);
    window:SetSize(320, 90);
    window:SetText("Importing into “" .. self.canvas:GetText() .. "”");
    CenterWindow(window);
    window:SetZOrder(2147483647);
        
    local label = Turbine.UI.Label();
    label:SetParent(window);
    label:SetFont(Turbine.UI.Lotro.Font.Verdana14);
    label:SetText("Paste exported data here (Ctrl+V):");
    label:SetSize(300, 16);
    label:SetPosition(10, 45);
    
    local textBox = Turbine.UI.Lotro.TextBox();
    textBox.canvas = self.canvas;
    textBox:SetParent(window);
    textBox:SetPosition(10, 61);
    textBox:SetSize(300, 20);
    textBox.TextChanged = function(t)
        textBox:SetWantsUpdates(true);
    end
    textBox.Update = function(t)
        t:SetWantsUpdates(false);
        t.TextChanged = nil;
        t:GetParent():Close();
        self.canvas:ImportShapes(t:GetText());
    end
    
    window:Activate();
    textBox:Focus();
    self.importWindow = window;
end


function Editor:CreateTextShape()
    local shape = TextShape();
    shape:SetID(self.canvas:GetNextShapeID());
    shape:SetOwner(self.playerName);
    self.canvas:AddShapeAt(shape, 50, 50);
    shape:SetForeColor(self.buttonTray:GetForeColor());
    shape:SetText("Some text");
    shape:AutoSize();
    return shape;
end

function Editor:CreateImageShape()
    local shape = ImageShape();
    shape:SetID(self.canvas:GetNextShapeID());
    shape:SetOwner(self.playerName);
    self.canvas:AddShapeAt(shape, 50, 50);
    shape:SetBackground(0x411BB710);
    shape:AutoSize();
    return shape;
end

function Editor:CreateRectangleShape()
    local shape = RectangleShape();
    shape:SetID(self.canvas:GetNextShapeID());
    shape:SetOwner(self.playerName);
    self.canvas:AddShapeAt(shape, 50, 50);
    shape:SetOutlineSize(self.buttonTray:GetThickness());
    shape:SetForeColor(self.buttonTray:GetForeColor());
    shape:SetBackColor(self.buttonTray:GetBackColor());
    return shape;
end

function Editor:CreateRadarMapShape()
    local shape = RadarMapShape();
    shape:SetID(self.canvas:GetNextShapeID());
    shape:SetOwner(self.playerName);
    self.canvas:AddShapeAt(shape, 50, 50);
    return shape;
end

function Editor:Update()
    local currentTime = Turbine.Engine.GetGameTime();
    if (currentTime - self.idleTime < 0.25) then
        return;
    end
    -- Monitor canvas for changes
    if (self.sendButtonNeedsUpdate) then
--self:SetWantsUpdates(false);
        changeStr = self.canvas:GetChanges();
        self.sendButton:SetData(changeStr);
        self.shimmering:Add(self.sendButton);
        self.sendButtonNeedsUpdate = false;
    end
end

function Editor:SetCommand(command)
    self.sendButton:SetCommand(command);
end
