Encoder = class();

-- to do: error checking
-- to do: encode nil values as "_"
-- to do: make all values indexable with ">" prefix
--        auto-build index?
--        string = prefix + indexed + suffix?

function Encoder:Constructor(numbered, properties)
    self.encoding = { numbered = {}; properties = {} };
    self.decoding = { numbered = {}; properties = {} };

    local encodeFuncs = {
        String = function(token, value)
            return string.len(value) .. token .. value;
        end;

        Integer = function(token, value)
            return value .. token;
        end;
        
        IntegerList = function(token, value)
            local prev, str = 0, "";
            for _, n in ipairs(value) do
                local delta = n - prev;
                str = str .. (((delta >= 0) and "+") or "-") .. math.abs(delta);
                prev = n;
            end
            --return String(token, str);
            return string.len(str) .. token .. str;
        end;

        Fractional = function(token, value)
            if ((value == 0) or (value == 1)) then
                return Integer(token, value);
            end
            local str = "." .. string.match(tostring(value), "^0[,%.]([0-9]+)$");
            return str .. token;
        end;

        Boolean = function(token, value)
            return (value and "1" or "0") .. token;
        end;

        Color = function(token, value)
            local color = Thurallor.Utils.Color(value.A, value.R, value.G, value.B);
            local str = color:GetHex();
            if (value.A ~= 1) then
                str = string.format("%2.2X", math.floor(value.A * 255 + 0.5)) .. str;
            end
            --return String(token, str);
            return string.len(str) .. token .. str;
        end;

        AssetID = function(token, value)
            return (value - 0x41000000) .. token;
        end;

        Font = function(token, value)
            return (value - 1107296256) .. token;
        end;

        Indexed = function(token, value)
            return self.encoding.numbered[value] .. token;
        end;
    };
    
    local decodeFuncs = {
        String = function(length, str)
            return str:sub(1, length), str:sub(length + 1);
        end;

        Integer = function(number, more)
            return tonumber(number), more;
        end;

        IntegerList = function(length, str)
            local str, more = str:sub(1, length), str:sub(length + 1);
            local value, prev = {}, 0;
            for delta in string.gmatch(str, "([%+%-][0-9]+)") do
                n = prev + tonumber(delta);
                table.insert(value, n);
                prev = n;
            end
            return value, more;
        end;
        
        Fractional = function(number, more)
            return tonumber(number), more;
        end;

        Boolean = function(number, more)
            return (number ~= "0"), more;
        end;

        Color = function(length, str)
            --local hex, more = String(length, str);
            local hex, more = str:sub(1, length), str:sub(length + 1);
            local alpha = 1;
            if (length == 8) then
                alpha = tonumber("0x" .. hex:sub(1, 2)) / 255;
            end
            local color = Thurallor.Utils.Color(alpha, 0, 0, 0);
            return color:SetHex(hex:sub(-6, -1)), more;
        end;

        AssetID = function(id, more)
            return 0x41000000 + id, more;
        end;

        Font = function(id, more)
            return 1107296256 + id, more;
        end;

        Indexed = function(index, more)
            return self.decoding.numbered[tonumber(index)], more;
        end;
    };

    -- Save the provided encoding tables, and invert them to create decoding tables
    for index, value in pairs(numbered) do
        self.encoding.numbered[value] = index;
        self.decoding.numbered[index] = value;
    end
    for name, info in pairs(properties) do
        local token, valueType, repeatable = unpack(info);
        self.encoding.properties[name] = { token, encodeFuncs[valueType], repeatable };
        self.decoding.properties[token] =  { name, decodeFuncs[valueType], repeatable };
    end
end

function Encoder:Encode(properties)
--Puts("Encoding: properties = " .. Serialize(properties, 1));
    local str = "";
    for name, value in pairs(properties) do
--Puts("   encoding: " .. name .. " = " .. Serialize(value, 1));
        local token, encodeFunc, repeatable = unpack(self.encoding.properties[name]);
        if (not token) then
            error("Unrecognized property: " .. name);
        end
        if (repeatable and (type(value) == "table")) then
            for _, v in ipairs(value) do
                str = str .. encodeFunc(token, v);
            end
        else
            str = str .. encodeFunc(token, value);
        end
    end
--Puts("Finished encoding; str = " .. str);
    return str;
end

function Encoder:Decode(str)
--Puts("Decoding: " .. str);
    local properties = {};
    local number, token, more, name, decodeFunc, value;
    repeat
        number, token, more = string.match(str, "^([0-9%.%-]+)(%a)(.*)$");
----Puts("number, token, more = " .. Serialize({number, token, more}));        
        name, decodeFunc, repeatable = unpack(self.decoding.properties[token]);
        if (not name) then
            error("Unrecognized token: " .. token);
        end
        value, str = decodeFunc(number, more);
--Puts("   decoding: " .. name .. " = " .. Serialize(value, 1));
        if (repeatable) then
            if (not properties[name]) then
                properties[name] = {};
            end
            table.insert(properties[name], value);
        else
            properties[name] = value;
        end
    until (str == "");
--Puts("Finished decoding; properties = " .. Serialize(properties, 1));
    return properties;
end

-- A single instance of the encoder is used by all Shape-derived objects
local numbered = {
    Shape,
    Group,
    Canvas,
    ImageShape,
    TextShape,
    RectangleShape,
    RadarMapShape
};
local properties = { -- token, type, allow multiple values
    BackColor =   { "C", "Color"       , false };
    Background =  { "b", "AssetID"     , false };
    Backgrounds = { "B", "IntegerList" , false };
    Class =       { "x", "Indexed"     , false };
    Deleted =     { "d", "Boolean"     , false };
    Font =        { "f", "Font"        , false };
    ForeColor =   { "c", "Color"       , false };
    Height =      { "h", "Integer"     , false };
    Hollow =      { "H", "Boolean"     , false };
    ID =          { "i", "Integer"     , false };
    Left =        { "l", "Integer"     , false };
    Location =    { "L", "String"      , false };
    Opacity =     { "o", "Fractional"  , false };
    OutlineSize = { "O", "Integer"     , false };
    Owner =       { "W", "String"      , false };
    PinEnabled =  { "p", "Boolean"     , false };
    Selected =    { "s", "Boolean"     , false };
    Shape =       { "S", "String"      , true  };
    Text =        { "T", "String"      , false };
    Thickness =   { "k", "Integer"     , false };
    Top =         { "t", "Integer"     , false };
    Version =     { "V", "Integer"     , false };
    Visible =     { "v", "Boolean"     , false };
    Width =       { "w", "Integer"     , false };
    XOffset =     { "X", "Integer"     , false };
    YOffset =     { "Y", "Integer"     , false };
    ZOrder =      { "z", "Integer"     , false };
};
Shape.encoder = Encoder(numbered, properties);
