PropertyTray = class(Turbine.UI.Control);

function PropertyTray:Constructor()
    Turbine.UI.Control.Constructor(self);
    self.editors = {};
    self.editorOrder = {};
    self.showingEditors = {};
    
    self.font = Turbine.UI.Lotro.Font.Verdana16;
    
    -- Create an editor for the Font property
    self:CreateFontEditor("Font");
    
    -- Create an editor for the Background property
    self:CreateAssetIDEditor("Background", "Image");

    -- Create an editor for the Hollow property
    self:CreateBooleanEditor("Hollow");

    -- Create an editor for the Location property
    self:CreateLocationEditor("Location");

    -- Create an editor for the PinEnabled property
    self:CreateBooleanEditor("PinEnabled", "Show pushpin");
end

function PropertyTray:CreateEditor(propertyName, labelText)
    local editor = Turbine.UI.Control();
    editor.propertyName = propertyName;
    editor.tooltipText = "(Tooltip missing)";
    editor:SetHeight(20);
    Thurallor.UI.Tooltip(function(e) return e.tooltipText; end):Attach(editor);

    local label = Turbine.UI.Label();
    label:SetFont(self.font);
    label:SetFontStyle(Turbine.UI.FontStyle.Outline);
    label:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleRight);
    label:SetText((labelText or propertyName) .. " ");
    label:AutoSize();
    label:SetHeight(20);
    label:SetParent(editor);
    label:SetMouseVisible(false);
    editor.label = label;
    
    local valueContainer = Turbine.UI.Control();
    valueContainer:SetHeight(20);
    valueContainer:SetParent(editor);
    valueContainer:SetMouseVisible(false);
    editor.valueContainer = valueContainer;

    function editor.SetWidth(e, width)
        Turbine.UI.Control.SetWidth(e, width);
        local space = width - e.valueContainer:GetWidth() - e.label:GetWidth();
        local left = math.floor(0.5 + space / 2);
        e.label:SetLeft(left);
        e.valueContainer:SetLeft(left + e.label:GetWidth());
    end
    
    self.editors[propertyName] = editor;
    table.insert(self.editorOrder, propertyName);
    return editor;
end

function PropertyTray:CreateStringEditor(propertyName, textBoxWidth, labelText)
    local editor = self:CreateEditor(propertyName, labelText);
   
    local textBox = Turbine.UI.Lotro.TextBox();
    textBox:SetFont(self.font);
    textBox:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleLeft);
    textBox:SetHeight(20);
    textBox:SetParent(editor.valueContainer);
    editor.textBox = textBox;
    textBox.editor = editor;
    textBox.TextChanged = function(t)
        DoCallbacks(t.editor, "ValueChanged");
    end

    function editor.valueContainer.SetWidth(v, width)
        Turbine.UI.Control.SetWidth(v, width);
        v:GetParent().textBox:SetWidth(width);
    end
    editor.valueContainer:SetWidth(textBoxWidth);
    
    function editor.SetValue(e, str)
        e.textBox:SetText(str or "");
    end

    function editor.GetValue(e)
        return e.textBox:GetText();
    end
    
    return editor;
end

function PropertyTray:CreateLocationEditor(propertyName, labelText)
    local textBoxWidth = 380;
    local editor = self:CreateStringEditor(propertyName, textBoxWidth, labelText);

-- to do: don't forget to localize:
    local cmd = "/loc";
    local button = Thurallor.UI.TrickButton(cmd, Turbine.ChatType.Standard, "^You are at: (.*)\n");
    button:SetParent(editor.valueContainer);
    button:SetFont(Turbine.UI.Lotro.Font.Verdana14);
    button:SetText("Detect");
    button:SetWidth(80);
    editor.button = button;
    button.editor = editor;
    button.Click = function(b, locStr)
        b.editor:SetValue(locStr);
        DoCallbacks(b.editor, "ValueChanged");
    end
    
    function editor.valueContainer.SetWidth(v, width)
        Turbine.UI.Control.SetWidth(v, width);
        local editor = v:GetParent();
        local textBoxWidth = width - editor.button:GetWidth();
        editor.textBox:SetWidth(textBoxWidth);
        editor.button:SetLeft(textBoxWidth);
    end
    editor.valueContainer:SetWidth(textBoxWidth + button:GetWidth());
end


function PropertyTray:CreateAssetIDEditor(propertyName, labelText)
    local editor = self:CreateStringEditor("Background", 250, labelText);

    function editor.SetValue(e, value)
        if (type(value) == "string") then
            value = tonumber(value) or value;
        end
        if (type(value) == "number") then
            value = string.format("0x%8.8X", value);
        end
        e.textBox:SetText(value or "");
    end
end

function PropertyTray:CreateListEditor(propertyName, strings, dropDownWidth, labelText)
    local editor = self:CreateEditor(propertyName, labelText);

    local dropDown = Thurallor.UI.DropDown(strings, strings[1]);
    dropDown:SetFont(self.font);
    dropDown:SetAlignment(Turbine.UI.ContentAlignment.MiddleLeft);
    dropDown:SetHeight(20);
    dropDown:SetParent(editor.valueContainer);
    dropDown.values = values;
    --Thurallor.UI.Tooltip(function(d) return d:GetParent().tooltipText; end):Attach(dropDown);
    editor.dropDown = dropDown;
    dropDown.editor = editor;
    dropDown.SelectionChanged = function(d, args)
        DoCallbacks(d.editor, "ValueChanged");
    end;

    function editor.valueContainer.SetWidth(v, width)
        Turbine.UI.Control.SetWidth(v, width);
        v:GetParent().dropDown:SetWidth(width);
    end
    editor.valueContainer:SetWidth(dropDownWidth);

    return editor;
end

function PropertyTray:CreateFontEditor(propertyName, labelText)
    local fontNames = {};
    for name, value in pairs(Turbine.UI.Lotro.Font) do
        if (name ~= "IsA") then
            table.insert(fontNames, name);
        end
    end
    table.sort(fontNames);
    local editor = self:CreateListEditor("Font", fontNames, 150, labelText);

    function editor.SetValue(e, font)
        for name, value in pairs(Turbine.UI.Lotro.Font) do
            if (value == font) then
                e.dropDown:SetSelectedItem(name);
            end
        end
    end
    
    function editor.GetValue(e)
        local fontName = e.dropDown:GetSelectedItem();
        return Turbine.UI.Lotro.Font[fontName];
    end

    return editor;
end

function PropertyTray:CreateBooleanEditor(propertyName, labelText)
    local editor = self:CreateEditor(propertyName, labelText);

    local checkBox = Turbine.UI.Lotro.CheckBox();
    checkBox:SetSize(16, 20);
    checkBox:SetParent(editor.valueContainer);
    --Thurallor.UI.Tooltip(function(t) return t:GetParent().tooltipText; end):Attach(textBox);
    editor.checkBox = checkBox;
    checkBox.editor = editor;
    checkBox.CheckedChanged = function(c)
        DoCallbacks(c.editor, "ValueChanged");
    end

    function editor.SetValue(e, flag)
        e.checkBox:SetChecked(flag);
    end

    function editor.GetValue(e)
        return e.checkBox:IsChecked();
    end

    return editor;
end


function PropertyTray:SetSize(width, height)
    Turbine.UI.Control.SetSize(self, width, height);
    self:Redraw();
end

function PropertyTray:SetContext(context, shapes)
--Puts("PropertyTray:SetContext " .. Serialize({context, shapeTypes}));
    self.context = context;
    self.showingEditors = {};
    if (context == "Global") then
    elseif (context == "Shape") then
        local shapeClass = next(shapes);
        if (next(shapes, shapeClass)) then
            -- Multiple types of shape selected.  Hide all buttons.
            -- To do: Handle selection of multiple shape types: show only the properties that apply to all of them.
        else
            if (#shapes[shapeClass] > 1) then
                -- Multiple shapes of this type selected.  Hide all buttons.
                -- To do: Handle selection of multiple shapes of the same type
            else
                self.shape = shapes[shapeClass][1];
                if (shapeClass == TextShape) then
                    self.showingEditors.Font = true;
                    self.editors.Font.tooltipText = "Change the text font for the selected shape";
                elseif (shapeClass == ImageShape) then
                    self.showingEditors.Background = true;
                    self.editors.Background.tooltipText = "Specify a file or the ID of a built-in image asset";
                elseif (shapeClass == RectangleShape) then
                    self.showingEditors.Hollow = true;
                    self.editors.Hollow.tooltipText = "Specify whether the interior of the shape is filled or transparent";
                elseif (shapeClass == RadarMapShape) then
                    self.showingEditors.Location = true;
                    self.editors.Location.tooltipText = "Specify the center point for the map (from the “/loc” chat command)";
                    self.showingEditors.PinEnabled = true;
                    self.editors.PinEnabled.tooltipText = "Show/hide the green pushpin";
                end
            end
        end
    end
    self:Redraw();
end

function PropertyTray:SetShape(shapes)
    -- currently only accepts a single shape
    self.shape = shapes[1];
    self:Redraw();
end

function PropertyTray:Redraw()
    self:GetControls():Clear();
    local width, height = self:GetSize();

    local editor;
    local spacing, top = 1, 0;
    for _, name in ipairs(self.editorOrder) do
        if (self.showingEditors[name]) then
            local editor = self.editors[name];
            editor:SetParent(self);
            editor:SetWidth(width);
            editor:SetPosition(0, top);
            editor:SetValue(self.shape:GetProperty(name));
            editor.ValueChanged = function(e)
                local setFunc = self.shape["Set" .. e.propertyName];
                setFunc(self.shape, e:GetValue());
            end
            top = top + editor:GetHeight() + spacing;
        end
    end
end
