-- This class provides a "main object" with generic plugin services:
--  Encapsulates the Turbine.Plugin and Turbine.PluginData objects.
--  Automatically loads setup at startup, and saves them when the user is idle.
--  Handles data reformatting to work around the save/load localization bug.

-- Events:
--  Loaded - generated after all source files have been parsed and executed
--  Unloading - generated before the plugin is unloaded



Plugin = class(Turbine.UI.Window);

-- This object should be constructed during the plugin's loading phase.  Pass the _G.plugin variable as the parameter.
function Plugin:Constructor(plugin)
    Turbine.UI.Window.Constructor(self);
    self.plugin = plugin;
    self.name = plugin:GetName();
    self.author = plugin:GetAuthor();
    self.version = plugin:GetVersion();
    for _, data in pairs(Turbine.PluginManager:GetAvailablePlugins()) do
        if (data.Name == self.name) then
            self.description = data.Description;
            self.image = data.Image;
            self.package = data.Package;
            self.apartment = data.ScriptState;
            break;
        end
    end

    self.runState = "loading";
    self.fileName = string.gsub(self.name, " ", "");
    self:LoadSettings();

    plugin.Load = function()
        self.runState = "running";
        DoCallbacks(self, "Loaded");
    end
    plugin.Unload = function()
        self.runState = "unloading";
        DoCallbacks(self, "Unloading");
        self:Close();
    end
end

function Plugin:LoadSettings()
    self.settings = {
        [Turbine.DataScope.Account] =   ImportTable(Turbine.PluginData.Load(Turbine.DataScope.Account,   self.fileName) or {});
        [Turbine.DataScope.Character] = ImportTable(Turbine.PluginData.Load(Turbine.DataScope.Character, self.fileName) or {});
        [Turbine.DataScope.Server] =    ImportTable(Turbine.PluginData.Load(Turbine.DataScope.Server,    self.fileName) or {});
    };
end

function Plugin:GetSettings(dataScope)
    return self.settings[dataScope];
end

function Plugin:SaveSettings(dataScope, immediately)
-- to do: If 'immediately' is false, schedule the save to occur during idle time.
    Turbine.PluginData.Save(dataScope, self.fileName, ExporTable(self.settings[dataScope]));
end

function Plugin:GetName()
    return self.name;
end

function Plugin:GetAuthor()
    return self.author;
end

function Plugin:GetDescription()
    return self.description;
end

function Plugin:GetImage()
    return self.image;
end

function Plugin:GetVersion()
    return self.version;
end

function Plugin:GetApartment()
    return self.apartment;
end

function Plugin:GetPackage()
    return self.package;
end

Thurallor = Thurallor or {};
Thurallor.Utils = Thurallor.Utils or {};
Thurallor.Utils.Plugin = Plugin;
